# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2020-02-26 16:13:25 UTC
import datetime

version = "3.2.2"
githash = "ce42e6e238c200a4715785ef8c9d233f612d0c75"


major = 3
minor = 2
bugfix = 2

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2020, 2, 26, 16, 13, 25)
debug = False

astropy_helpers_version = ""
