# Autogenerated by Astropy-affiliated package ligo.skymap's setup.py on 2020-02-26 16:13:25 UTC
import datetime

version = "0.1.16"
githash = "546e05aad0b3905d7705d260f9b76b4df5104b9d"


major = 0
minor = 1
bugfix = 16

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2020, 2, 26, 16, 13, 25)
debug = False

astropy_helpers_version = "3.2.2"
